#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from typing import Union, Optional

import torch
import numpy as np
from plum import dispatch

from .core import is_less_then_epsilon_4th_root, arc_sin_x_over_x
from .conversion import quaternion_from_matrix, quaternion_from_axis_angle, quaternion_from_euler_angle, axis_angle_from_quaternion, \
    matrix_from_quaternion, euler_angle_from_quaternion


# Created by fzhiheng on 2024/1/15
# Copyright (c) 2024 fzhiheng. All rights reserved.
# 2024/1/15 下午7:27


@dispatch
def weighted_average_quaternions(quaternions: np.ndarray, weights: Optional[np.ndarray] = None):
    """

    Args:
        quaternions: is a (*,N,4) numpy matrix and contains the quaternions to average in the rows.
            The quaternions are arranged as (w,x,y,z), with w being the scalar
        weights: The weight of the quaternions (*,N,1)

    Returns:the average quaternion of the input. Note that the signs of the output quaternion can be reversed,
        since q and -q describe the same orientation # (*,4)

    Raises: ValueError if all weights are zero

    """

    # Number of quaternions to average
    if weights is None:
        weights = np.ones_like(quaternions[..., 0:1])

    mat_a = quaternions.swapaxes(-1, -2) @ (quaternions * weights)  # (*,4,4)
    weight_sum = np.sum(weights, axis=-2, keepdims=True)  # (*,1,1)
    if np.any(weight_sum <= 0.0):
        raise ValueError("At least one weight must be greater than zero")

    # scale
    mat_a = (1.0 / weight_sum) * mat_a

    eigen_values, eigen_vectors = np.linalg.eig(mat_a)

    index = np.argmax(eigen_values, axis=-1)
    eigen_vectors = np.take_along_axis(eigen_vectors, index[..., None, None], axis=-1)  # (*,4,1)
    eigen_vectors = np.squeeze(eigen_vectors, axis=-1)  # (*,4)

    return np.real((eigen_vectors))


@dispatch
def weighted_average_quaternions(quaternions: torch.Tensor, weights: Optional[torch.Tensor] = None):
    """

    Args:
        quaternions: is a (*,N,4)  matrix, w,x,y,z
        weights:

    Returns:# (*,4) (*,4)

    """

    if weights is None:
        weights = torch.ones_like(quaternions[..., 0:1])

    weight_sum = torch.sum(weights, dim=-2, keepdim=True)  # (*,1,1)
    if torch.any(weight_sum <= 0.0):
        raise ValueError("At least one weight must be greater than zero")

    mat_a = torch.transpose(quaternions, -1, -2) @ (quaternions * weights)  # (*,4,4)
    mat_a = (1.0 / weight_sum) * mat_a
    eigen_values, eigen_vectors = torch.linalg.eig(mat_a)  # (*,4), # (*,4,4)
    eigen_values = torch.real(eigen_values)
    index = torch.argmax(eigen_values, dim=-1)

    eigen_vectors = torch.take_along_dim(eigen_vectors, index[..., None, None], dim=-1)  # (*,4,1)
    eigen_vectors = torch.squeeze(eigen_vectors, dim=-1)  # (*,4)
    return torch.real((eigen_vectors))


class Quaternion(object):
    def __init__(self, quat: torch.Tensor):
        """

        Args:
            quat: (*,4) wxyz
        """
        self.data = quat
        self.check()
        self.normalize()

    def check(self):
        if self.data.dtype != torch.float32 and self.data.dtype != torch.float64:
            raise TypeError("Quaternion must be float32 or float64")
        if self.data.shape[-1] != 4:
            raise ValueError("Quaternion must be wxyz")

    @property
    def shape(self):
        return self.data.shape

    @property
    def w(self):
        return self.data[..., 0:1]

    @property
    def x(self):
        return self.data[..., 1:2]

    @property
    def y(self):
        return self.data[..., 2:3]

    @property
    def z(self):
        return self.data[..., 3:4]

    def __getitem__(self, item):
        if isinstance(item, tuple) and len(item) == self.data.ndim:
            if isinstance(item[-1], slice):
                if not (item[-1].start is None and item[-1].stop is None and item[-1].step is None):
                    raise IndexError("Last dimension of Quaternion cannot be sliced")
            else:
                raise IndexError("Last dimension of Quaternion cannot be sliced")
        return Quaternion(self.data[item])

    def __add__(self, other):
        return Quaternion(self.data + other.data)

    def __sub__(self, other):
        return Quaternion(self.data - other.data)

    def __mul__(self, other: 'Quaternion') -> 'Quaternion':
        """
        Args:
            other: (*,4) wxyz

        Returns: (*,4) wxyz

        """
        w1, x1, y1, z1 = self.w, self.x, self.y, self.z
        w2, x2, y2, z2 = other.w, other.x, other.y, other.z
        w3 = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
        x3 = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
        y3 = w1 * y2 - x1 * z2 + y1 * w2 + z1 * x2
        z3 = w1 * z2 + x1 * y2 - y1 * x2 + z1 * w2
        return Quaternion(torch.cat([w3, x3, y3, z3], dim=-1))

    def __div__(self, factor: Union[int, float, torch.Tensor]) -> 'Quaternion':
        return Quaternion(self.data / factor)

    __truediv__ = __div__

    def __hash__(self):
        return hash(self.data)

    def __eq__(self, other):
        return self.data == other.data

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return "{} data:\n{}".format(type(self).__name__, str(self.data))

    def conjugate(self):
        return Quaternion(torch.cat([self.w, -self.x, -self.y, -self.z], dim=-1))

    def norm(self) -> torch.Tensor:
        return torch.linalg.norm(self.data, dim=-1, keepdim=True)

    def normalize(self):
        self.data = self.data / self.norm()

    def inverse(self):
        return self.conjugate() / self.norm()

    def log(self) -> torch.Tensor:
        """

        Returns: (*,3) xyz

        """
        q_imagi = torch.cat([self.x, self.y, self.z], dim=-1)
        na = torch.linalg.norm(q_imagi, dim=-1, keepdim=True)
        eta = self.w
        scale1 = torch.where(eta >= 0, torch.acos(eta) / na, -torch.acos(-eta) / na)
        scale2 = torch.where(eta > 0, arc_sin_x_over_x(na), -arc_sin_x_over_x(na))
        scale = torch.where(torch.abs(eta) < na, scale1, scale2)
        return q_imagi * (2.0 * scale)

    @staticmethod
    def exp(dx: torch.Tensor) -> 'Quaternion':
        input_dtype = dx.dtype
        assert input_dtype == torch.float32 or input_dtype == torch.float64

        theta = torch.linalg.norm(dx, dim=-1, keepdim=True)
        one_over_48 = 1.0 / 48.0
        na1 = 0.5 + (theta * theta) * one_over_48
        na2 = torch.sin(theta * 0.5) / theta
        na = torch.where(is_less_then_epsilon_4th_root(theta), na1, na2)
        ct = torch.cos(theta * 0.5)
        wxyz = torch.cat([ct, dx * na], dim=-1)
        return Quaternion(wxyz)

    def interpolate(self, other: 'Quaternion', ratio: Union[float, torch.Tensor]) -> 'Quaternion':
        """

        Args:
            other: another Quaternion
            ratio: interpolation ratio

        Returns: interpolated Quaternion

        """
        # check the range of ratio
        if isinstance(ratio, float):
            if ratio < 0 or ratio > 1:
                raise ValueError(f"ratio must be in [0,1], but get {ratio}")
        elif isinstance(ratio, torch.Tensor):
            if (ratio < 0).any() or (ratio > 1).any():
                raise ValueError(f"ratio must be in [0,1], but get {ratio}")
        else:
            raise TypeError(f"ratio must be float or torch.Tensor, but get {type(ratio)}")

        q1_2 = self.conjugate() * other
        q_inter = self * Quaternion.exp(ratio * q1_2.log())
        return q_inter

    @staticmethod
    def average(quats: 'Quaternion', weights: Optional[torch.Tensor] = None) -> 'Quaternion':
        """

        Args:
            quats: (*,N,4)  wxyz
            weights: (*,N)

        Returns:# (*,4)

        """
        average_quat = weighted_average_quaternions(quats.data, weights)
        return Quaternion(average_quat)

    @classmethod
    def from_matrix(cls, matrix: torch.Tensor) -> 'Quaternion':
        """

        Args:
            matrix: (*,3,3)

        Returns: (*,4) wxyz

        """
        return cls(quaternion_from_matrix(matrix))

    @classmethod
    def from_axis_angle(cls, axis_angle: torch.Tensor) -> 'Quaternion':
        """

        Args:
            axis_angle: (*,3)

        Returns: (*,4) wxyz

        """
        return cls(quaternion_from_axis_angle(axis_angle))

    @classmethod
    def from_euler_angle(cls, euler_angle: torch.Tensor) -> 'Quaternion':
        """

        Args:
            euler_angle: (*,3)

        Returns: (*,4) wxyz

        """
        return cls(quaternion_from_euler_angle(euler_angle))

    def to_matrix(self) -> torch.Tensor:
        """

        Returns: (*,3,3)

        """
        return matrix_from_quaternion(self.data)

    def to_axis_angle(self) -> torch.Tensor:
        """

        Returns: (*,3)

        """
        return axis_angle_from_quaternion(self.data)

    def to_euler_angle(self) -> torch.Tensor:
        """

        Returns: (*,3)

        """
        return euler_angle_from_quaternion(self.data)


if __name__ == "__main__":
    wxyz1 = torch.rand(2, 3, 4)
    wxyz2 = torch.rand(2, 3, 4)
    q1 = Quaternion(wxyz1)
    q2 = Quaternion(wxyz2)
    q3 = q1 * q2
    print(q1)
    print(q2)
    print(q3)
    print(q1.interpolate(q2, 0.5))



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from .quaternions import Quaternion, weighted_average_quaternions

# __init__.py.py
# Created by fzhiheng on 2024/1/16
# Copyright (c) 2024 fzhiheng. All rights reserved.
# 2024/1/16 上午9:34

__all__ = ['Quaternion', 'conversion']



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import torch
import numpy as np

from src.pyrotation.conversion import *

# convert_test.py
# Created by fzhiheng on 2024/1/16
# Copyright (c) 2024 fzhiheng. All rights reserved.
# 2024/1/16 下午5:10

if __name__ == "__main__":

    shape = (1, 1, 3)
    axis_angle = np.random.random(shape)

    matrix = matrix_from_axis_angle(axis_angle)
    quat = quaternion_from_axis_angle(axis_angle)
    euler = euler_angle_from_matrix(matrix)

    # check all functions
    assert np.allclose(axis_angle, axis_angle_from_quaternion(quat))
    assert np.allclose(axis_angle, axis_angle_from_matrix(matrix))
    assert np.allclose(axis_angle, axis_angle_from_euler(euler))

    assert np.allclose(matrix, matrix_from_quaternion(quat))
    assert np.allclose(matrix, matrix_from_euler_angle(euler))

    assert np.allclose(euler, euler_angle_from_quaternion(quat))
    assert np.allclose(euler, euler_angle_from_axis_angle(axis_angle))

    assert np.allclose(quat, quaternion_from_matrix(matrix))
    assert np.allclose(quat, quaternion_from_euler_angle(euler))

    axis_angle = torch.randn(shape)
    matrix = matrix_from_axis_angle(axis_angle)
    quat = quaternion_from_axis_angle(axis_angle)
    euler = euler_angle_from_matrix(matrix)

    # check all functions
    assert torch.allclose(axis_angle, axis_angle_from_quaternion(quat))
    assert torch.allclose(axis_angle, axis_angle_from_matrix(matrix))
    assert torch.allclose(axis_angle, axis_angle_from_euler(euler))

    assert torch.allclose(matrix, matrix_from_quaternion(quat))
    assert torch.allclose(matrix, matrix_from_euler_angle(euler))

    assert torch.allclose(euler, euler_angle_from_quaternion(quat))
    assert torch.allclose(euler, euler_angle_from_axis_angle(axis_angle))

    assert torch.allclose(quat, quaternion_from_matrix(matrix))
    assert torch.allclose(quat, quaternion_from_euler_angle(euler))

    se3 = np.random.rand(2, 2, 6)
    assert np.allclose(se3, se3_from_SE3(SE3_from_se3(se3)))

    se3 = torch.randn(2, 2, 6)
    assert torch.allclose(se3, se3_from_SE3(SE3_from_se3(se3)))

    print("test success!")





# -*- coding: UTF-8 -*-
from typing import Union

import torch
import numpy as np
from plum import dispatch

from .core import skew_symmetric, is_less_then_epsilon_4th_root, get_matrix_y, get_matrix_x, get_matrix_z

# File    ：quaternions.py
# Author  ：fzhiheng
# Date    ：2023/10/11 下午8:29

"""
This module contains functions for working with conversion between different rotational representations.
All functions support batched inputs and outputs. All functions support numpy arrays and torch tensors as input and output.

The following representations are supported:
    quaternions are represented as (w,x,y,z) with w being the scalar.
    matrix are represented as 3x3 matrices.
    axis angle are represented as (x,y,z) with theta being the angle of rotation around the axis.
    euler angles are represented as (yaw,pitch,roll), and yaw is the rotation around the z axis, pitch around the y axis and roll around the x axis.

"""


@dispatch
def check_matrix(matrix: np.ndarray):
    """
    Check if a matrix is a valid rotation matrix in right hand coordinate system.
    Raises a ValueError if this is not the case.

    Args:
        matrix: The matrix to check

    Returns: None

    Raises: ValueError if the matrix is not a valid rotation matrix

    """
    if not np.allclose(np.eye(3), matrix @ np.swapaxes(matrix, -1, -2), rtol=0, atol=1e-6):
        raise ValueError("The matrix is not a valid rotation matrix")
    # 检查是否是右手坐标系
    if not np.allclose(np.linalg.det(matrix), 1.0, rtol=0, atol=1e-6):
        raise ValueError("The matrix is in a right hand coordinate system")


@dispatch
def check_matrix(matrix: torch.Tensor):
    """
    Check if a matrix is a valid rotation matrix in right hand coordinate system.
    Raises a ValueError if this is not the case.

    Args:
        matrix: The matrix to check

    Returns: None

    Raises: ValueError if the matrix is not a valid rotation matrix

    """
    device = matrix.device
    if not torch.allclose(torch.eye(3, dtype=matrix.dtype).to(device), matrix @ matrix.transpose(-1, -2), rtol=0, atol=1e-6):
        raise ValueError("The matrix is not a valid rotation matrix")
    # 检查是否是右手坐标系
    if not torch.allclose(torch.det(matrix), torch.ones(1, dtype=matrix.dtype).to(device), rtol=0, atol=1e-6):
        raise ValueError("The matrix is in a right hand coordinate system")


# 旋转矩阵转四元数
@dispatch
def quaternion_from_matrix(matrix: np.ndarray) -> np.ndarray:
    """

    Args:
        matrix: (*,3,3)

    Returns: (*,4)

    """
    check_matrix(matrix)
    w = np.sqrt(1 + matrix[..., 0, 0] + matrix[..., 1, 1] + matrix[..., 2, 2]) / 2
    x = (matrix[..., 2, 1] - matrix[..., 1, 2]) / (4 * w)
    y = (matrix[..., 0, 2] - matrix[..., 2, 0]) / (4 * w)
    z = (matrix[..., 1, 0] - matrix[..., 0, 1]) / (4 * w)
    return np.stack([w, x, y, z], axis=-1)


# 旋转矩阵转四元数
@dispatch
def quaternion_from_matrix(matrix: torch.Tensor) -> torch.Tensor:
    """

    Args:
        matrix: (*,3,3)

    Returns: (*,4) wxyz

    """
    check_matrix(matrix)
    w = torch.sqrt(1 + matrix[..., 0, 0] + matrix[..., 1, 1] + matrix[..., 2, 2]) / 2  # (*,)
    x = (matrix[..., 2, 1] - matrix[..., 1, 2]) / (4 * w)  # (*,)
    y = (matrix[..., 0, 2] - matrix[..., 2, 0]) / (4 * w)  # (*,)
    z = (matrix[..., 1, 0] - matrix[..., 0, 1]) / (4 * w)  # (*,)
    return torch.stack([w, x, y, z], dim=-1)  # (*,4)


# 轴角转四元数
@dispatch
def quaternion_from_axis_angle(axis_angle: np.ndarray) -> np.ndarray:
    """

    Args:
        axis_angle: (*,3)

    Returns: wxyz

    """
    angle = np.linalg.norm(axis_angle, axis=-1, keepdims=True)  # (*,1)
    axis = axis_angle / angle  # (*,3)
    sin_result = np.sin(angle / 2)  # (*,1)
    return np.concatenate([np.cos(angle / 2), axis * sin_result], axis=-1)  # (*,4)


# 轴角转四元数
@dispatch
def quaternion_from_axis_angle(axis_angle: torch.Tensor) -> torch.Tensor:
    """

    Args:
        axis_angle: (*,3)

    Returns: wxyz

    """
    angle = torch.linalg.norm(axis_angle, dim=-1, keepdim=True)
    axis = axis_angle / angle
    sin_result = torch.sin(angle / 2)
    return torch.cat([torch.cos(angle / 2), axis * sin_result], dim=-1)


# 欧拉角转四元数
@dispatch
def quaternion_from_euler_angle(euler: np.ndarray) -> np.ndarray:
    """

    Args:
        euler: (*,3) 默认是ypr,内旋

    Returns: wxyz

    """
    euler = euler / 2
    yaw, pitch, roll = euler[..., 0], euler[..., 1], euler[..., 2]  # (*,)
    w = np.sin(yaw) * np.sin(pitch) * np.sin(roll) + np.cos(yaw) * np.cos(pitch) * np.cos(roll)  # (*,)
    x = -np.sin(yaw) * np.sin(pitch) * np.cos(roll) + np.cos(yaw) * np.cos(pitch) * np.sin(roll)  # (*,)
    y = np.sin(yaw) * np.cos(pitch) * np.sin(roll) + np.cos(yaw) * np.sin(pitch) * np.cos(roll)  # (*,)
    z = -np.cos(yaw) * np.sin(pitch) * np.sin(roll) + np.sin(yaw) * np.cos(pitch) * np.cos(roll)
    return np.stack([w, x, y, z], axis=-1)  # (*,4)


# 欧拉角转四元数
@dispatch
def quaternion_from_euler_angle(euler: torch.Tensor) -> torch.Tensor:
    """

    Args:
        euler: (*,3) 默认是ypr,内旋

    Returns: wxyz

    """
    euler = euler / 2
    yaw, pitch, roll = euler[..., 0], euler[..., 1], euler[..., 2]
    w = torch.sin(yaw) * torch.sin(pitch) * torch.sin(roll) + torch.cos(yaw) * torch.cos(pitch) * torch.cos(roll)
    x = -torch.sin(yaw) * torch.sin(pitch) * torch.cos(roll) + torch.cos(yaw) * torch.cos(pitch) * torch.sin(roll)
    y = torch.sin(yaw) * torch.cos(pitch) * torch.sin(roll) + torch.cos(yaw) * torch.sin(pitch) * torch.cos(roll)
    z = -torch.cos(yaw) * torch.sin(pitch) * torch.sin(roll) + torch.sin(yaw) * torch.cos(pitch) * torch.cos(roll)
    return torch.stack([w, x, y, z], dim=-1)


# 四元数转旋转矩阵
@dispatch
def matrix_from_quaternion(quaternion: np.ndarray) -> np.ndarray:
    """

    Args:
        quaternion: (*,4) w,x,y,z

    Returns: rotation_matrix (*,3,3)

    """
    w, x, y, z = quaternion[..., 0], quaternion[..., 1], quaternion[..., 2], quaternion[..., 3]
    rot_matx = np.zeros((*quaternion.shape[:-1], 3, 3))
    rot_matx[..., 0, 0] = np.power(w, 2) + np.power(x, 2) - np.power(y, 2) - np.power(z, 2)
    rot_matx[..., 0, 1] = 2 * x * y - 2 * w * z
    rot_matx[..., 0, 2] = 2 * x * z + 2 * w * y
    rot_matx[..., 1, 0] = 2 * x * y + 2 * w * z
    rot_matx[..., 1, 1] = np.power(w, 2) - np.power(x, 2) + np.power(y, 2) - np.power(z, 2)
    rot_matx[..., 1, 2] = 2 * y * z - 2 * w * x
    rot_matx[..., 2, 0] = 2 * x * z - 2 * w * y
    rot_matx[..., 2, 1] = 2 * y * z + 2 * w * x
    rot_matx[..., 2, 2] = np.power(w, 2) - np.power(x, 2) - np.power(y, 2) + np.power(z, 2)
    return rot_matx


# 四元数转旋转矩阵
@dispatch
def matrix_from_quaternion(quaternion: torch.Tensor) -> torch.Tensor:
    """

    Args:
        quaternion: (*,4) wxyz

    Returns: rotation_matrix (*,3,3)

    """
    w, x, y, z = quaternion[..., 0], quaternion[..., 1], quaternion[..., 2], quaternion[..., 3]
    rot_matx = torch.zeros((*quaternion.shape[:-1], 3, 3), device=quaternion.device)
    rot_matx[..., 0, 0] = torch.pow(w, 2) + torch.pow(x, 2) - torch.pow(y, 2) - torch.pow(z, 2)
    rot_matx[..., 0, 1] = 2 * x * y - 2 * w * z
    rot_matx[..., 0, 2] = 2 * x * z + 2 * w * y
    rot_matx[..., 1, 0] = 2 * x * y + 2 * w * z
    rot_matx[..., 1, 1] = torch.pow(w, 2) - torch.pow(x, 2) + torch.pow(y, 2) - torch.pow(z, 2)
    rot_matx[..., 1, 2] = 2 * y * z - 2 * w * x
    rot_matx[..., 2, 0] = 2 * x * z - 2 * w * y
    rot_matx[..., 2, 1] = 2 * y * z + 2 * w * x
    rot_matx[..., 2, 2] = torch.pow(w, 2) - torch.pow(x, 2) - torch.pow(y, 2) + torch.pow(z, 2)
    return rot_matx


def matrix_from_euler_angle(euler_angle: Union[torch.Tensor, np.ndarray], axes=("z", "y", "x")) -> Union[torch.Tensor, np.ndarray]:
    """ get rotation matrix from euler angle,default order is yaw, pitch, roll

    Args:
        euler_angle (Union[torch.Tensor, np.ndarray]): (*,3), (yaw, pitch, roll)

    Returns:
            Union[torch.Tensor, np.ndarray]: (*,3,3)
    """
    yaw = euler_angle[..., 0]
    pitch = euler_angle[..., 1]
    roll = euler_angle[..., 2]
    R_z = get_matrix_z(yaw)
    R_y = get_matrix_y(pitch)
    R_x = get_matrix_x(roll)
    axis_rotation = {"z": R_z, "y": R_y, "x": R_x}
    matrix = axis_rotation[axes[0]] @ axis_rotation[axes[1]] @ axis_rotation[axes[2]]
    return matrix


# 轴角转旋转矩阵
@dispatch
def matrix_from_axis_angle(axis_angle: np.ndarray) -> np.ndarray:
    """

    Args:
        axis_angle: (*,3)

    Returns: (*,3,3)

    """
    angle = np.linalg.norm(axis_angle, axis=-1, keepdims=True)  # (*,1)
    axis = axis_angle / angle  # (*,3)
    # https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
    axis_skew = skew_symmetric(axis)  # (*,3,3)
    angle_cos = np.cos(angle)[..., None]  # (*,1,1)
    angle_sin = np.sin(angle)[..., None]  # (*,1,1)

    R = angle_cos * np.eye(3) + angle_sin * axis_skew + (1 - angle_cos) * axis[..., None] @ axis[..., None, :]

    return R


@dispatch
def matrix_from_axis_angle(axis_angle: torch.Tensor) -> torch.Tensor:
    """

    Args:
        axis_angle: (*,3)

    Returns: (*,3,3)

    """
    device = axis_angle.device
    angle = torch.linalg.norm(axis_angle, dim=-1, keepdim=True)
    axis = axis_angle / angle
    # 使用罗德里格旋转公式
    # https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
    axis_skew = skew_symmetric(axis)
    angle_cos = torch.cos(angle)[..., None]
    angle_sin = torch.sin(angle)[..., None]

    R = angle_cos * torch.eye(3).to(device) + angle_sin * axis_skew + (1 - angle_cos) * axis[..., None] @ axis[..., None, :]
    return R


# 四元数转轴角
@dispatch
def axis_angle_from_quaternion(quaternion: np.ndarray) -> np.ndarray:
    """

    Args:
        quaternion: (*,4)

    Returns: (*,3)

    """
    angle = 2 * np.arccos(quaternion[..., 0:1])
    axis = quaternion[..., 1:] / np.sin(angle / 2)
    return angle * axis


@dispatch
def axis_angle_from_quaternion(quaternion: torch.Tensor) -> torch.Tensor:
    """四元数转轴角

    Args:
        quaternion: (*,4)

    Returns: (*,3)

    """
    angle = 2 * torch.arccos(quaternion[..., 0:1])
    axis = quaternion[..., 1:] / torch.sin(angle / 2)
    return angle * axis


# 旋转矩阵转轴角
@dispatch
def axis_angle_from_matrix(matrix: np.ndarray) -> np.ndarray:
    """旋转矩阵转轴角

    Args:
        matrix: (*,3,3)

    Returns: (*,3)

    """
    check_matrix(matrix)
    angle = np.arccos((matrix[..., 0, 0] + matrix[..., 1, 1] + matrix[..., 2, 2] - 1) / 2)  # (*,)
    angle = angle[..., None]
    axis = np.stack(
        [
            matrix[..., 2, 1] - matrix[..., 1, 2],
            matrix[..., 0, 2] - matrix[..., 2, 0],
            matrix[..., 1, 0] - matrix[..., 0, 1],
        ],
        axis=-1,
    )
    axis = axis / (2 * np.sin(angle))
    return angle * axis


@dispatch
def axis_angle_from_matrix(matrix: torch.Tensor) -> torch.Tensor:
    """旋转矩阵转轴角

    Args:
        matrix: (*,3,3)

    Returns: (*,3)

    """
    check_matrix(matrix)
    angle = torch.arccos((matrix[..., 0, 0] + matrix[..., 1, 1] + matrix[..., 2, 2] - 1) / 2)  # (*,)
    angle = angle[..., None]
    axis = torch.stack(
        [
            matrix[..., 2, 1] - matrix[..., 1, 2],
            matrix[..., 0, 2] - matrix[..., 2, 0],
            matrix[..., 1, 0] - matrix[..., 0, 1],
        ],
        dim=-1,
    )  # (*,3)
    axis = axis / (2 * torch.sin(angle))
    return angle * axis


# 欧拉角转轴角
def axis_angle_from_euler(euler_angle: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
    """欧拉角转轴角

    Args:
        euler_angle: (*,3)

    Returns: (*,3)

    """
    matrix = matrix_from_euler_angle(euler_angle)
    axis_angle = axis_angle_from_matrix(matrix)
    return axis_angle


# 四元数转欧拉角
@dispatch
def euler_angle_from_quaternion(quaternion: np.ndarray) -> np.ndarray:
    """四元数转欧拉角
    https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles
    Args:
        quaternion: (*,4)

    Returns: (*,3)

    """
    w, x, y, z = quaternion[..., 0], quaternion[..., 1], quaternion[..., 2], quaternion[..., 3]
    # roll (x-axis rotation)
    sinr_cosp = 2 * (w * x + y * z)
    cosr_cosp = 1 - 2 * (x * x + y * y)
    roll = np.arctan2(sinr_cosp, cosr_cosp)
    # pitch (y-axis rotation)
    sinp = 2 * (w * y - z * x)
    pitch = np.where(np.abs(sinp) >= 1, np.sign(sinp) * np.pi / 2, np.arcsin(sinp))
    # yaw (z-axis rotation)
    siny_cosp = 2 * (w * z + x * y)
    cosy_cosp = 1 - 2 * (y * y + z * z)
    yaw = np.arctan2(siny_cosp, cosy_cosp)
    return np.stack([yaw, pitch, roll], axis=-1)


@dispatch
def euler_angle_from_quaternion(quaternion: torch.Tensor) -> torch.Tensor:
    """四元数转欧拉角
    https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles
    Args:
        quaternion: (*,4)

    Returns: (*,3)

    """
    w, x, y, z = quaternion[..., 0], quaternion[..., 1], quaternion[..., 2], quaternion[..., 3]
    # roll (x-axis rotation)
    sinr_cosp = 2 * (w * x + y * z)
    cosr_cosp = 1 - 2 * (x * x + y * y)
    roll = torch.atan2(sinr_cosp, cosr_cosp)
    # pitch (y-axis rotation)
    sinp = 2 * (w * y - z * x)
    pitch = torch.where(torch.abs(sinp) >= 1, torch.sign(sinp) * np.pi / 2, torch.arcsin(sinp))
    # yaw (z-axis rotation)
    siny_cosp = 2 * (w * z + x * y)
    cosy_cosp = 1 - 2 * (y * y + z * z)
    yaw = torch.atan2(siny_cosp, cosy_cosp)
    return torch.stack([yaw, pitch, roll], dim=-1)


# 旋转矩阵转欧拉角
@dispatch
def euler_angle_from_matrix(matrix: np.ndarray) -> np.ndarray:
    """旋转矩阵转欧拉角
    https://www.learnopencv.com/rotation-matrix-to-euler-angles/
    Args:
        matrix: (*,3,3)

    Returns: (*,3)

    """
    check_matrix(matrix)
    sy = np.sqrt(matrix[..., 0, 0] * matrix[..., 0, 0] + matrix[..., 1, 0] * matrix[..., 1, 0])
    singular = sy < 1e-6
    roll = np.where(singular, np.arctan2(-matrix[..., 1, 2], matrix[..., 1, 1]),
                    np.arctan2(matrix[..., 2, 1], matrix[..., 2, 2]))
    pitch = np.where(singular, np.arctan2(-matrix[..., 2, 0], sy), np.arctan2(-matrix[..., 2, 0], sy))
    yaw = np.where(singular, 0, np.arctan2(matrix[..., 1, 0], matrix[..., 0, 0]))
    return np.stack([yaw, pitch, roll], axis=-1)


@dispatch
def euler_angle_from_matrix(matrix: torch.Tensor) -> torch.Tensor:
    """旋转矩阵转欧拉角
    https://www.learnopencv.com/rotation-matrix-to-euler-angles/
    Args:
        matrix: (*,3,3)

    Returns: (*,3)

    """
    check_matrix(matrix)
    sy = torch.sqrt(matrix[..., 0, 0] * matrix[..., 0, 0] + matrix[..., 1, 0] * matrix[..., 1, 0])
    singular = sy < 1e-6
    roll = torch.where(singular, torch.arctan2(-matrix[..., 1, 2], matrix[..., 1, 1]),
                       torch.arctan2(matrix[..., 2, 1], matrix[..., 2, 2]))
    pitch = torch.where(singular, torch.arctan2(-matrix[..., 2, 0], sy), torch.arctan2(-matrix[..., 2, 0], sy))
    yaw = torch.where(singular, 0, torch.arctan2(matrix[..., 1, 0], matrix[..., 0, 0]))
    return torch.stack([yaw, pitch, roll], dim=-1)


# 轴角转欧拉角
def axis_angle_from_euler_angle(euler_angle: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
    """欧拉角转轴角
    Args:
        euler_angle: (*,3)

    Returns: (*,3)

    """
    return axis_angle_from_quaternion(quaternion_from_euler_angle(euler_angle))


def euler_angle_from_axis_angle(axis_angle: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
    """轴角转欧拉角
    Args:
        axis_angle: (*,3)

    Returns: (*,3)

    """
    return euler_angle_from_quaternion(quaternion_from_axis_angle(axis_angle))


# other functions
@dispatch
def fill_matrix(matrix: np.ndarray, t: np.ndarray) -> np.ndarray:
    """_summary_

    Args:
        matrix (np.ndarray): (*,3,3) matrix
        t (np.ndarray): (*,3)

    Returns:
        np.ndarray: (*,4,4)
    """
    full_matrix = np.concatenate([matrix, t[..., None]], axis=-1)
    padding = np.zeros_like(full_matrix[..., :1, :])
    padding[..., 0, -1] = 1
    full_matrix = np.concatenate([full_matrix, padding], axis=-2)
    return full_matrix


@dispatch
def fill_matrix(matrix: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
    """_summary_

    Args:
        matrix (torch.Tensor): (*,3,3) matrix
        t (torch.Tensor): (*,3)

    Returns:
        torch.Tensor: (*,4,4)
    """
    full_matrix = torch.cat([matrix, t[..., None]], dim=-1)
    padding = torch.zeros_like(full_matrix[..., :1, :])
    padding[..., 0, -1] = 1
    full_matrix = torch.cat([full_matrix, padding], dim=-2)
    return full_matrix


def full_matrix_from_qt(q: Union[np.ndarray, torch.Tensor], t: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
    """_summary_

    Args:
        q (Union[np.ndarray, torch.Tensor]): (*,4) wxyz
        t (Union[np.ndarray, torch.Tensor]): (*,3)

    Returns:
        Union[np.ndarray, torch.Tensor]: (*,4,4)
    """
    matrix = matrix_from_quaternion(q)  # (*,3,3)
    full_matrix = fill_matrix(matrix, t)
    return full_matrix


def SO3_from_so3(phi: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
    """_summary_

    Args:
        so3 (Union[np.ndarray, torch.Tensor]): (*,3)

    Returns:
        Union[np.ndarray, torch.Tensor]: (*,3,3)
    """
    return matrix_from_axis_angle(phi)


def so3_from_SO3(SO3: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
    """_summary_

    Args:
        SO3 (Union[np.ndarray, torch.Tensor]): (*,3,3)

    Returns:
        Union[np.ndarray, torch.Tensor]: (*,3)
    """
    return axis_angle_from_matrix(SO3)


@dispatch
def SE3_from_se3(se3: np.ndarray) -> np.ndarray:
    """_summary_

    Args:
        se3 (Union[np.ndarray, torch.Tensor]): (*,6) (phi, rho) phi: rotation, rho: translation

    Returns:
        Union[np.ndarray, torch.Tensor]: (*,4,4)
    """
    phi = se3[..., :3]  # (*,3)
    rho = se3[..., 3:]  # (*,3)
    angle = np.linalg.norm(phi, axis=-1, keepdims=True)
    a = phi / angle
    skew = skew_symmetric(a)
    angle_cos = np.cos(angle)
    angle_sin = np.sin(angle)
    R = angle_cos[..., None] * np.eye(3) + angle_sin[..., None] * skew + (1 - angle_cos[..., None]) * a[..., None] @ a[..., None, :]
    tmp1 = np.where(is_less_then_epsilon_4th_root(np.abs(angle)), 1.0 - angle * angle * (1.0 / 6.0), angle_sin / angle)
    tmp2 = np.where(is_less_then_epsilon_4th_root(np.abs(angle)), angle * (1.0 / 2.0) - angle * angle * angle * (1.0 / 24.0), (1 - angle_cos) / angle)
    tmp1 = tmp1[..., None]
    tmp2 = tmp2[..., None]
    jacobi = tmp1 * np.eye(3) + (1 - tmp1) * a[..., None] @ a[..., None, :] + tmp2 * skew  # (*,3,3)
    t = jacobi @ rho[..., None]
    SE3 = fill_matrix(R, t[..., 0])
    return SE3


@dispatch
def SE3_from_se3(se3: torch.Tensor) -> torch.Tensor:
    """_summary_

    Args:
        se3 (Union[np.ndarray, torch.Tensor]): (*,6) (phi, rho) phi: rotation, rho: translation

    Returns:
        Union[np.ndarray, torch.Tensor]: (*,4,4)
    """
    phi = se3[..., :3]  # (*,3)
    rho = se3[..., 3:]  # (*,3)
    device = se3.device
    angle = torch.linalg.norm(phi, dim=-1, keepdim=True)
    a = phi / angle
    skew = skew_symmetric(a)
    angle_cos = torch.cos(angle)
    angle_sin = torch.sin(angle)
    R = angle_cos[..., None] * torch.eye(3).to(device) + angle_sin[..., None] * skew + (1 - angle_cos[..., None]) * a[..., None] @ a[..., None, :]
    tmp1 = torch.where(is_less_then_epsilon_4th_root(torch.abs(angle)), 1.0 - angle * angle * (1.0 / 6.0), angle_sin / angle)
    tmp2 = torch.where(is_less_then_epsilon_4th_root(torch.abs(angle)), angle * (1.0 / 2.0) - angle * angle * angle * (1.0 / 24.0),
                       (1 - angle_cos) / angle)
    tmp1 = tmp1[..., None]
    tmp2 = tmp2[..., None]
    jacobi = tmp1 * torch.eye(3).to(device) + (1 - tmp1) * a[..., None] @ a[..., None, :] + tmp2 * skew  # (*,3,3)
    t = jacobi @ rho[..., None]
    SE3 = fill_matrix(R, t[..., 0])
    return SE3


@dispatch
def se3_from_SE3(SE3: np.ndarray) -> np.ndarray:
    """

    Args:
        SE3: (*,4,4)

    Returns:

    """
    R = SE3[..., :3, :3]
    t = SE3[..., :3, 3:4]
    phi = so3_from_SO3(R)  # (*,3)
    angle = np.linalg.norm(phi, axis=-1, keepdims=True)
    skew = skew_symmetric(phi)  # (*,3,3)
    angle_cos = np.cos(angle)
    angle_sin = np.sin(angle)
    tmp1 = np.where(is_less_then_epsilon_4th_root(np.abs(angle)), 1.0 - angle * angle * (1.0 / 6.0), angle_sin / angle)
    tmp2 = np.where(is_less_then_epsilon_4th_root(np.abs(angle)), 0.5 - angle * angle * (1.0 / 24.0), (1 - angle_cos) / angle ** 2)
    tmp = (1 - tmp1 / (2 * tmp2)) / (angle ** 2)

    jacobi_inv = np.ones_like(tmp[..., None]) * np.eye(3) - 0.5 * skew + tmp[..., None] * skew @ skew  # (*,3,3)
    rho = (jacobi_inv @ t)[..., 0]

    se3 = np.concatenate([phi, rho], -1)
    return se3


@dispatch
def se3_from_SE3(SE3: torch.Tensor) -> torch.Tensor:
    """

    Args:
        SE3: (*,4,4)

    Returns:

    """
    R = SE3[..., :3, :3]
    t = SE3[..., :3, 3:4]
    device = SE3.device
    phi = so3_from_SO3(R)  # (*,3)
    angle = torch.linalg.norm(phi, axis=-1, keepdims=True)
    skew = skew_symmetric(phi)  # (*,3,3)
    angle_cos = torch.cos(angle)
    angle_sin = torch.sin(angle)
    tmp1 = torch.where(is_less_then_epsilon_4th_root(np.abs(angle)), 1.0 - angle * angle * (1.0 / 6.0), angle_sin / angle)
    tmp2 = torch.where(is_less_then_epsilon_4th_root(np.abs(angle)), 0.5 - angle * angle * (1.0 / 24.0), (1 - angle_cos) / angle ** 2)
    tmp = (1 - tmp1 / (2 * tmp2)) / (angle ** 2)

    jacobi_inv = torch.ones_like(tmp[..., None]).to(device) * torch.eye(3, dtype=SE3.dtype).to(device) - 0.5 * skew + tmp[..., None] * skew @ skew
    rho = (jacobi_inv @ t)[..., 0]

    se3 = torch.cat([phi, rho], -1)
    return se3



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import torch
import numpy as np

from plum import dispatch


# core.py
# Created by fzhiheng on 2024/1/15
# Copyright (c) 2024 fzhiheng. All rights reserved.
# 2024/1/15 下午4:02

@dispatch
def skew_symmetric(vector: np.ndarray):
    """

    Args:
        vector:  (*,3)

    Returns: (*,3,3)

    """
    skew1 = np.stack([np.zeros_like(vector[..., 0]), -vector[..., 2], vector[..., 1]], axis=-1)  # (*,3)
    skew2 = np.stack([vector[..., 2], np.zeros_like(vector[..., 0]), -vector[..., 0]], axis=-1)  # (*,3)
    skew3 = np.stack([-vector[..., 1], vector[..., 0], np.zeros_like(vector[..., 0])], axis=-1)  # (*,3)
    skew = np.stack([skew1, skew2, skew3], axis=-2)  # (*,3,3)
    return skew


@dispatch
def skew_symmetric(vector: torch.Tensor):
    """

    Args:
        vector:  (*,3)

    Returns: (*,3,3)

    """
    device = vector.device
    skew1 = torch.stack([torch.zeros_like(vector[..., 0]).to(device), -vector[..., 2], vector[..., 1]], dim=-1)
    skew2 = torch.stack([vector[..., 2], torch.zeros_like(vector[..., 0]).to(device), -vector[..., 0]], dim=-1)
    skew3 = torch.stack([-vector[..., 1], vector[..., 0], torch.zeros_like(vector[..., 0]).to(device)], dim=-1)
    skew = torch.stack([skew1, skew2, skew3], dim=-2)
    return skew

@dispatch
def get_matrix_z(yaw: np.ndarray) -> np.ndarray:
    """ roation matrix around z-axis

    Args:
        yaw (np.ndarray): (*,)

    Returns:
        np.ndarray: (*,3,3)
    """
    Rz = np.stack(
        [
            np.stack([np.cos(yaw), -np.sin(yaw), np.zeros_like(yaw)], axis=-1),
            np.stack([np.sin(yaw), np.cos(yaw), np.zeros_like(yaw)], axis=-1),
            np.stack([np.zeros_like(yaw), np.zeros_like(yaw), np.ones_like(yaw)], axis=-1),
        ],
        axis=-2,
    )

    return Rz


@dispatch
def get_matrix_y(pitch: np.ndarray) -> np.ndarray:
    """ roation matrix around y-axis

    Args:
        pitch (np.ndarray): (*,)

    Returns:
        np.ndarray: (*,3,3)
    """
    Ry = np.stack(
        [
            np.stack([np.cos(pitch), np.zeros_like(pitch), np.sin(pitch)], axis=-1),
            np.stack([np.zeros_like(pitch), np.ones_like(pitch), np.zeros_like(pitch)], axis=-1),
            np.stack([-np.sin(pitch), np.zeros_like(pitch), np.cos(pitch)], axis=-1),
        ],
        axis=-2,
    )

    return Ry


@dispatch
def get_matrix_x(roll: np.ndarray) -> np.ndarray:
    """ roation matrix around x-axis

    Args:
        roll (np.ndarray): (*,)
    Returns:
        np.ndarray: (*,3,3)
    """

    Rx = np.stack(
        [
            np.stack([np.ones_like(roll), np.zeros_like(roll), np.zeros_like(roll)], axis=-1),
            np.stack([np.zeros_like(roll), np.cos(roll), -np.sin(roll)], axis=-1),
            np.stack([np.zeros_like(roll), np.sin(roll), np.cos(roll)], axis=-1),
        ],
        axis=-2,
    )
    return Rx


@dispatch
def get_matrix_z(yaw: torch.Tensor) -> torch.Tensor:
    """ roation matrix around z-axis

    Args:
        yaw (torch.Tensor): (*,)
    Returns:
        torch.Tensor: (*,3,3)
    """
    device = yaw.device
    Rz = torch.stack(
        [
            torch.stack([torch.cos(yaw), -torch.sin(yaw), torch.zeros_like(yaw).to(device)], dim=-1),
            torch.stack([torch.sin(yaw), torch.cos(yaw), torch.zeros_like(yaw).to(device)], dim=-1),
            torch.stack([torch.zeros_like(yaw).to(device),
                         torch.zeros_like(yaw).to(device),
                         torch.ones_like(yaw).to(device)], dim=-1),
        ],
        dim=-2,
    )
    return Rz


@dispatch
def get_matrix_y(pitch: torch.Tensor) -> torch.Tensor:
    """ roation matrix around y-axis

    Args:
        pitch (torch.Tensor): (*,)
    Returns:
        torch.Tensor: (*,3,3)
    """
    device = pitch.device
    Ry = torch.stack(
        [
            torch.stack([torch.cos(pitch), torch.zeros_like(pitch).to(device), torch.sin(pitch)], dim=-1),
            torch.stack([torch.zeros_like(pitch).to(device),
                         torch.ones_like(pitch).to(device),
                         torch.zeros_like(pitch).to(device)], dim=-1),
            torch.stack(
                [-torch.sin(pitch), torch.zeros_like(pitch).to(device), torch.cos(pitch)], dim=-1),
        ],
        dim=-2,
    )
    return Ry


@dispatch
def get_matrix_x(roll: torch.Tensor) -> torch.Tensor:
    """ roation matrix around x-axis

    Args:
        roll (torch.Tensor): (*,)
    Returns:
        torch.Tensor: (*,3,3)
    """
    device = roll.device
    Rx = torch.stack(
        [
            torch.stack([torch.ones_like(roll).to(device),
                         torch.zeros_like(roll).to(device),
                         torch.zeros_like(roll).to(device)], dim=-1),
            torch.stack([torch.zeros_like(roll).to(device), torch.cos(roll), -torch.sin(roll)], dim=-1),
            torch.stack([torch.zeros_like(roll).to(device), torch.sin(roll), torch.cos(roll)], dim=-1),
        ],
        dim=-2,
    )
    return Rx


@dispatch
def is_less_then_epsilon_4th_root(x: np.ndarray):
    input_dtype = x.dtype
    assert input_dtype == np.float32 or input_dtype == np.float64
    return x < np.power(np.finfo(input_dtype).eps, 1.0 / 4.0)


@dispatch
def is_less_then_epsilon_4th_root(x: torch.Tensor):
    input_dtype = x.dtype
    device = x.device
    assert input_dtype == torch.float32 or input_dtype == torch.float64
    return x < torch.pow(torch.finfo(input_dtype).eps, torch.tensor(1.0 / 4.0, dtype=input_dtype, device=device))


@dispatch
def arc_sin_x_over_x(x: np.ndarray) -> np.ndarray:
    """

    Args:
        x:  (*,1)

    Returns: (*,1)

    """
    return np.where(is_less_then_epsilon_4th_root(np.abs(x)), 1.0 + x * x * (1.0 / 6.0), np.arcsin(x) / x)


@dispatch
def arc_sin_x_over_x(x: torch.Tensor) -> torch.Tensor:
    """

    Args:
        x:  (*,1)

    Returns: (*,1)

    """
    return torch.where(is_less_then_epsilon_4th_root(torch.abs(x)), 1.0 + x * x * (1.0 / 6.0), torch.asin(x) / x)


@dispatch
def sin_x_over_x(x: np.ndarray) -> np.ndarray:
    return np.where(is_less_then_epsilon_4th_root(np.abs(x)), 1.0 - x * x * (1.0 / 6.0), np.sin(x) / x)


@dispatch
def sin_x_over_x(x: torch.Tensor) -> torch.Tensor:
    return torch.where(is_less_then_epsilon_4th_root(torch.abs(x)), 1.0 - x * x * (1.0 / 6.0), torch.sin(x) / x)
